@@ -1,13 +1,19 @@
-__author__ = "Andrew Rechnitzer"
-__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer"
-__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
-__license__ = "AGPLv3"
-import os
-import toml
+# SPDX-License-Identifier: AGPL-3.0-or-later
+# Copyright (C) 2018-2020 Andrew Rechnitzer
+# Copyright (C) 2018 Elvis Cai
+# Copyright (C) 2019-2021 Colin B. Macdonald
+# Copyright (C) 2020 Victoria Schuster
+# Copyright (C) 2020 Vala Vakilian
+
 import re
 import time
+import logging
+from pathlib import Path
 
-from PyQt5.QtCore import Qt, pyqtSignal, QSize, QTimer
+import toml
+import appdirs
+
+from PyQt5.QtCore import Qt, pyqtSignal, QTimer
 from PyQt5.QtGui import QDropEvent, QStandardItem, QStandardItemModel
 from PyQt5.QtWidgets import (
     QAbstractItemView,
@@ -19,22 +25,26 @@ from PyQt5.QtWidgets import (
     QFormLayout,
     QGridLayout,
     QItemDelegate,
-    QMessageBox,
     QPushButton,
     QSpinBox,
     QTableView,
     QTextEdit,
     QLineEdit,
-    QToolButton,
     QVBoxLayout,
     QWidget,
 )
 
 
-def commentLoadAll():
-    """Grab comments from the toml file or return defaults."""
+log = logging.getLogger("annotr")
+comment_dir = Path(appdirs.user_data_dir("plom", "PlomGrading.org"))
 
-    clist_defaults = """
+
+def comments_new_default_list():
+    """Make a default list of comments."""
+
+    # string-escaping here must match toml.dumps
+    cdict = toml.loads(
+        r"""
 [[comment]]
 delta = -1
 text = "algebra"
@@ -49,7 +59,7 @@ text = "meh"
 
 [[comment]]
 delta = 0
-text = 'tex: you can write \LaTeX, $e^{i\pi}+1=0$'
+text = "tex: you can write \\LaTeX, $e^{i\\pi}+1=0$"
 
 [[comment]]
 delta = 0
@@ -70,6 +80,23 @@ delta = -1
 text = "Quest. 2 specific comment"
 tags = "Q2 foo bar"
 """
+    )
+    # should be a dict = {"comment": [list of stuff]}
+    assert "comment" in cdict
+    clist = cdict["comment"]
+    return comments_apply_default_fields(clist)
+
+
+def comments_apply_default_fields(comlist):
+    """Add missing fields with defaults to list of comments.
+
+    Args:
+        comlist (list): list of dicts.  Copies will not be made so
+            keep a deep copy if you need the original.
+
+    Returns:
+        list: updated list of dicts.
+    """
     comment_defaults = {
         "tags": "",
         "testname": "",
@@ -77,25 +104,61 @@ tags = "Q2 foo bar"
         "created": time.gmtime(0),
         "modified": time.gmtime(0),
     }
-    # TODO: don't save empty tags/testnames/etc to file
-    if os.path.exists("plomComments.toml"):
-        cdict = toml.load("plomComments.toml")
-    else:
-        cdict = toml.loads(clist_defaults)
-    # should be a dict = {"comment": [list of stuff]}
-    assert "comment" in cdict
-    clist = cdict["comment"]
-    for d in clist:
+    for d in comlist:
         for k, v in comment_defaults.items():
             d.setdefault(k, comment_defaults[k])
+    return comlist
+
+
+def comments_load_from_file(f):
+    """Grab comments from a toml file.
+
+    Args:
+        f (str/pathlib.Path): filename of a toml file.
+
+    Returns:
+        list: list of dicts, one for each comments.
+
+    Raises:
+        FileNotFoundError:
+        PermissionError:
+    """
+    cdict = toml.load(f)
+    clist = cdict["comment"]
+    return comments_apply_default_fields(clist)
+
+
+def commentLoadAll():
+    """Grab comments from the toml file or return defaults."""
+    local_comfile = Path("plomComments.toml")
+    comfile = comment_dir / "plomComments.toml"
+    try:
+        clist = comments_load_from_file(local_comfile)
+        # Note: on save, central file overwritten, Issue #1355
+        log.info("Loaded a LOCAL comment file: %s", local_comfile)
+        return clist
+    except (FileNotFoundError, PermissionError):
+        pass
+    try:
+        clist = comments_load_from_file(comfile)
+        log.info("Loaded comment file: %s", comfile)
+        return clist
+    except FileNotFoundError:
+        pass
+    clist = comments_new_default_list()
+    log.info("Starting from scratch (no comment file %s)", comfile)
     return clist
 
 
-def commentSaveList(clist):
+def comments_save_list(clist, comment_dir=comment_dir, filename="plomComments.toml"):
     """Export comment list to toml file."""
-    # toml wants a dictionary
-    with open("plomComments.toml", "w") as fname:
+    # TODO: don't save empty tags/testnames/etc to file?
+    comfile = comment_dir / filename
+    comment_dir.mkdir(exist_ok=True)
+    with open(comfile, "w") as fname:
+        # toml wants a dictionary
         toml.dump({"comment": clist}, fname)
+    log.info("Saved comment file: %s", comfile)
 
 
 # Eventually there may be more "state" to the filters and something like a dict
@@ -117,7 +180,7 @@ def commentVisibleInQuestion(com, n):
     Qn = "Q{}".format(n)
     tags = com["tags"].split()
     return any([t == Qn for t in tags]) or not any(
-        [re.match("^Q\d+$", t) for t in tags]
+        [re.match(r"^Q\d+$", t) for t in tags]
     )
 
 
@@ -145,7 +208,7 @@ def commentTaggedQn(com, n):
 
 def commentHasMultipleQTags(com):
     tags = com["tags"].split()
-    x = [1 if re.match("^Q\d+$", t) else 0 for t in tags]
+    x = [1 if re.match(r"^Q\d+$", t) else 0 for t in tags]
     return sum(x) >= 2
 
 
@@ -160,12 +223,14 @@ class CommentWidget(QWidget):
         self.parent = parent
         self.maxMark = maxMark
         grid = QGridLayout()
+        # assume our container will deal with margins
+        grid.setContentsMargins(0, 0, 0, 0)
         # the table has 2 cols, delta&comment.
         self.CL = SimpleCommentTable(self)
         grid.addWidget(self.CL, 1, 1, 2, 3)
         self.addB = QPushButton("Add")
         self.delB = QPushButton("Delete")
-        self.filtB = QPushButton("Filter...")
+        self.filtB = QPushButton("Filter")
         grid.addWidget(self.addB, 3, 1)
         grid.addWidget(self.filtB, 3, 2)
         grid.addWidget(self.delB, 3, 3)
@@ -188,11 +253,6 @@ class CommentWidget(QWidget):
         self.questnum = n
         self.CL.populateTable()
 
-    def setQuestionNumberFromTGV(self, tgv):
-        """Extract question number from TGV, set, and refresh view."""
-        n = int(self.parent.tgv[5:7])
-        self.setQuestionNumber(n)
-
     def setStyle(self, markStyle):
         # The list needs a style-delegate because the display
         # of the delta-mark will change depending on
@@ -210,6 +270,13 @@ class CommentWidget(QWidget):
         self.CL.delegate.currentMark = currentMark
         self.CL.viewport().update()
 
+    def reset(self):
+        """Return the widget to a no-TGV-specified state."""
+        self.setQuestionNumber(None)
+        self.setTestname(None)
+        # TODO: do we need to do something about maxMark, currentMax, markStyle?
+        self.CL.populateTable()
+
     def saveComments(self):
         self.CL.saveCommentList()
 
@@ -241,7 +308,8 @@ class CommentWidget(QWidget):
         return self.CL.getCurrentItemRow()
 
     def setCurrentItemRow(self, r):
-        self.CL.selectRow(r)
+        """Reset the comment row on a new task to last highlighted comment."""
+        return self.CL.setCurrentItemRow(r)
 
     def addFromTextList(self):
         # text items in scene.
@@ -453,14 +521,20 @@ class SimpleCommentTable(QTableView):
             # Get the row on which to drop
             dropRow = self.drop_on(event)
             dropIdx = self.cmodel.index(dropRow, 2).data()
+            log.debug(
+                "debug DnD: drag={}, drop={}".format((row, idx), (dropRow, dropIdx))
+            )
             # If we drag from earlier row, handle index after deletion
             if row < dropRow:
                 dropRow -= 1
-            com = self.clist.pop(row)
+            # TODO: maybe `row` does not account for hidden rows, changed to `idx`
+            com = self.clist.pop(int(idx))
+            log.debug("debug DnD: com of row drag is {}".format(com))
             self.clist.insert(dropRow, com)
             self.populateTable()
 
             # Reselect the dropped row (TODO: does this work?)
+            # TODO: sometimes dropRow is None (drag to end?)
             self.selectRow(dropRow)
 
             # Resize the rows - they were expanding after drags for some reason
@@ -529,22 +603,22 @@ class SimpleCommentTable(QTableView):
         # the comment signal for the annotator to read.
         if index == 0:  # make sure something is selected
             self.currentItem()
-        r = self.selectedIndexes()[0].row()
-        self.commentSignal.emit(
-            [self.cmodel.index(r, 0).data(), self.cmodel.index(r, 1).data()]
-        )
+        r = self.getCurrentItemRow()
+        if r is not None:
+            self.commentSignal.emit(
+                [self.cmodel.index(r, 0).data(), self.cmodel.index(r, 1).data()]
+            )
 
     def saveCommentList(self):
-        commentSaveList(self.clist)
+        comments_save_list(self.clist)
 
     def deleteItem(self):
         # Remove the selected row (or do nothing if no selection)
-        sel = self.selectedIndexes()
-        if len(sel) == 0:
+        r = self.getCurrentItemRow()
+        if r is None:
             return
-        idx = int(self.cmodel.index(sel[0].row(), 2).data())
+        idx = int(self.cmodel.index(r, 2).data())
         self.clist.pop(idx)
-        # self.cmodel.removeRow(sel[0].row())
         # TODO: maybe sloppy to rebuild, need automatic cmodel ontop of clist
         self.populateTable()
 
@@ -559,34 +633,49 @@ class SimpleCommentTable(QTableView):
     def currentItem(self):
         # If no selected row, then select row 0.
         # else select current row - triggers a signal.
-        sel = self.selectedIndexes()
-        if len(sel) == 0:
-            self.selectRow(0)
-        else:
-            self.selectRow(sel[0].row())
+        r = self.getCurrentItemRow()
+        if r is None:
+            if self.cmodel.rowCount() >= 1:
+                r = 0
+        self.setCurrentItemRow(r)
 
     def getCurrentItemRow(self):
-        if self.selectedIndexes():
-            return self.selectedIndexes()[0].row()
+        """Return the currently-selected row or None if no selection."""
+        if not self.selectedIndexes():
+            return None
+        return self.selectedIndexes()[0].row()
 
     def setCurrentItemRow(self, r):
-        self.selectRow(r)
+        """Reset the comment row on a new task to last highlighted comment.
+
+        Args:
+            r (int): The integer representing the row number in the
+                comments table.  If r is None, do nothing.
+        """
+        if r is not None:
+            self.selectRow(r)
 
     def nextItem(self):
-        # Select next row (wraps around)
-        sel = self.selectedIndexes()
-        if len(sel) == 0:
-            self.selectRow(0)
-        else:
-            self.selectRow((sel[0].row() + 1) % self.cmodel.rowCount())
+        """Move selection to the next row, wrapping around if needed."""
+        r = self.getCurrentItemRow()
+        if r is None:
+            if self.cmodel.rowCount() >= 1:
+                r = 0
+            else:
+                return
+        r = (r + 1) % self.cmodel.rowCount()
+        self.setCurrentItemRow(r)
 
     def previousItem(self):
-        # Select previous row (wraps around)
-        sel = self.selectedIndexes()
-        if len(sel) == 0:
-            self.selectRow(0)
-        else:
-            self.selectRow((sel[0].row() - 1) % self.cmodel.rowCount())
+        """Move selection to the prevoous row, wrapping around if needed."""
+        r = self.getCurrentItemRow()
+        if r is None:
+            if self.cmodel.rowCount() >= 1:
+                r = 0
+            else:
+                return
+        r = (r - 1) % self.cmodel.rowCount()
+        self.setCurrentItemRow(r)
 
     def insertItem(self, com):
         self.clist.append(com)
