@@ -6,3 +6,6 @@ __copyright__ = "Copyright (C) 2020 Andrew Rechnitzer and Colin B. Macdonald"
 __credits__ = "The Plom Project Developers"
 __license__ = "AGPL-3.0-or-later"
 # SPDX-License-Identifier: AGPL-3.0-or-later
+
+from .examDB import PlomDB
+from .buildPlomDB import buildExamDatabaseFromSpec
