@@ -1,11 +1,11 @@
-# -*- coding: utf-8 -*-
+# SPDX-License-Identifier: AGPL-3.0-or-later
+# Copyright (C) 2020 Colin B. Macdonald
 
 """Utils concerning rules about data, like valid student numbers."""
 
-__author__ = "Colin B. Macdonald"
-__copyright__ = "Copyright (C) 2020 Colin B. Macdonald"
-__license__ = "AGPL-3.0-or-later"
-# SPDX-License-Identifier: AGPL-3.0-or-later
+
+StudentIDLength = 8
+
 
 def isValidUBCStudentNumber(n):
     """Is this a valid student number for UBC?
@@ -18,10 +18,11 @@ def isValidUBCStudentNumber(n):
         return False
     if sid < 0:
         return False
-    if len(str(n)) != 8:
+    if len(str(n)) != StudentIDLength:
         return False
     return True
 
+
 def censorStudentNumber(n):
     """Replace some parts of student number with astericks."""
     n = str(n)
@@ -29,12 +30,14 @@ def censorStudentNumber(n):
     assert len(n) == len(r)
     return r
 
+
 def censorStudentName(s):
     """Replace most of a student student name with astericks."""
     if len(s) <= 3:
-        r = s[0] + "*"*7
+        r = s[0] + "*" * 7
     else:
-        r = s[:3] + "*"*5
+        r = s[:3] + "*" * 5
     return r
 
+
 isValidStudentNumber = isValidUBCStudentNumber
